from .rapidata_client import (
    RapidataClient,

    DemographicSelection,
    LabelingSelection,
    ValidationSelection,
    ConditionalValidationSelection,
    CappedSelection,

    RapidataSettings, 
    TranslationBehaviourOptions,
    AlertOnFastResponse,
    TranslationBehaviour,
    FreeTextMinimumCharacters,
    NoShuffle,
    PlayVideoUntilTheEnd,
    CustomSetting,

    CountryFilter,
    LanguageFilter,
    UserScoreFilter,
    CampaignFilter,
    AgeGroup,
    Gender,

    CountryCodes,
    MediaAsset,
    TextAsset,
    MultiAsset,
    RapidataDataTypes,
    Box,
    )



from rapidata.api_client.models.simple_workflow_model import SimpleWorkflowModel
from rapidata.api_client.models.simple_workflow_model_blueprint import SimpleWorkflowModelBlueprint
from rapidata.api_client.models.transcription_rapid_blueprint import TranscriptionRapidBlueprint
from rapidata.rapidata_client.workflow._base_workflow import Workflow


class SelectWordsWorkflow(Workflow):
    """
    A workflow for select words tasks.

    This class represents a select words workflow where audio or video content
    is transcribed based on given instructions.

    Attributes:
        _instruction (str): The instruction for the select words task.

    Args:
        instruction (str): The instruction to be provided for the select words task.
    """

    def __init__(self, instruction: str):
        super().__init__(type="SimpleWorkflowConfig")
        self._instruction = instruction

    def _to_model(self) -> SimpleWorkflowModel:
        blueprint = TranscriptionRapidBlueprint(
            _t="TranscriptionBlueprint",
            title=self._instruction
        )

        return SimpleWorkflowModel(
            _t="SimpleWorkflow",
            blueprint=SimpleWorkflowModelBlueprint(blueprint)
        )



from typing import Any
from rapidata.rapidata_client.filter._base_filter import RapidataFilter
from rapidata.api_client.models.age_user_filter_model import AgeUserFilterModel
from rapidata.rapidata_client.filter.models.age_group import AgeGroup


class AgeFilter(RapidataFilter):
    """AgeFilter Class
    
    Can be used to filter who to target based on age groups.
    
    
    Args:
        age_groups (list[AgeGroup]): List of age groups to filter by."""

    def __init__(self, age_groups: list[AgeGroup]):
        self.age_groups = age_groups

    def _to_model(self):
        return AgeUserFilterModel(
            _t="AgeFilter",
            ageGroups=[age_group._to_backend_model() for age_group in self.age_groups],
        )



from typing import Any
from rapidata.rapidata_client.selection._base_selection import RapidataSelection
from rapidata.api_client.models.labeling_selection import (
    LabelingSelection as LabelingSelectionModel,
)


class LabelingSelection(RapidataSelection):
    """Labeling selection class.
    
    Decides how many actual datapoints you want to show per session.
    
    Args:
        amount (int): The amount of labeling rapids that will be shown per session.
    """

    def __init__(self, amount: int):
        self.amount = amount

    def _to_model(self) -> Any:
        return LabelingSelectionModel(_t="LabelingSelection", amount=self.amount)



"""Media Asset Module

Defines the MediaAsset class for handling media file paths within assets.
"""

import os
from io import BytesIO
from rapidata.rapidata_client.assets._base_asset import BaseAsset
import requests
import re
from PIL import Image

class MediaAsset(BaseAsset):
    """MediaAsset Class

    Represents a media asset by storing the file path.
    Supports local files and URLs for images, MP3, and MP4.

    Args:
        path (str): The file system path to the media asset.

    Raises:
        FileNotFoundError: If the provided file path does not exist.
    """

    ALLOWED_TYPES = [
        'image/', 
        'audio/mp3',      # MP3
        'video/mp4',       # MP4
    ]

    def __init__(self, path: str):
        """
        Initialize a MediaAsset instance.

        Args:
            path (str): The file system path to the media asset or a URL.

        Raises:
            FileNotFoundError: If the provided file path does not exist.
            ValueError: If media type is unsupported or duration exceeds 25 seconds.
        """
        if not isinstance(path, str):
            raise ValueError("Media must be a string, either a local file path or a URL")
        
        if re.match(r'^https?://', path):
            self.path = self.__get_media_bytes(path)
            self.name = path.split('/')[-1]
            self.name = self.__check_name_ending(self.name)
            return
        
        if not os.path.exists(path):
            raise FileNotFoundError(f"File not found: {path}")
        
        self.path: str | bytes = path
        self.name = path

    def get_image_dimension(self) -> tuple[int, int] | None:
        """
        Get the dimensions (width, height) of an image file.
        Returns None for non-image files or if dimensions can't be determined.
        """
        if not any(self.name.lower().endswith(ext) for ext in ('.jpg', '.jpeg', '.png', '.gif', '.webp')):
            return None
            
        try:
            if isinstance(self.path, bytes):
                img = Image.open(BytesIO(self.path))
            else:
                img = Image.open(self.path)
            return img.size
        except Exception:
            return None

    def set_custom_name(self, name: str) -> 'MediaAsset':
        """Set a custom name for the media asset (only works with URLs)."""
        if isinstance(self.path, bytes):
            self.name = self.__check_name_ending(name)
        else:
            raise ValueError("Custom name can only be set for URLs.")
        return self
    
    def __check_name_ending(self, name: str) -> str:
        """Check if the media path is valid."""
        if not name.endswith(('.jpg', '.jpeg', '.png', '.gif', '.mp3', '.mp4', '.webp')):
            print("Warning: Supported file types: jpg, jpeg, png, gif, mp3, mp4. Image might not be displayed correctly.")
            name = name + '.jpg'
        return name

    def __get_media_bytes(self, url: str) -> bytes:
        """
        Downloads media files from URL and validates type and duration.
        
        Args:
            url: URL of the media file
                
        Returns:
            bytes: Media data
            
        Raises:
            ValueError: If media type is unsupported or duration exceeds limit
            requests.exceptions.RequestException: If download fails
        """
        response = requests.get(url, stream=False)  # Don't stream, we need full file
        response.raise_for_status()

        content_type = response.headers.get('content-type', '').lower()
        
        # Validate content type
        if not any(content_type.startswith(t) for t in self.ALLOWED_TYPES):
            raise ValueError(
                f'URL does not point to an allowed media type.\n'
                f'Content-Type: {content_type}\n'
                f'Allowed types: {self.ALLOWED_TYPES}'
            )

        content = BytesIO(response.content)
        return content.getvalue()



from ._base_metadata import Metadata
from ._private_text_metadata import PrivateTextMetadata
from ._public_text_metadata import PublicTextMetadata
from ._prompt_metadata import PromptMetadata
from ._select_words_metadata import SelectWordsMetadata



from rapidata.rapidata_client.assets import MediaAsset, TextAsset, MultiAsset
from rapidata.rapidata_client.metadata import Metadata
from typing import Sequence
from rapidata.rapidata_client.validation.rapids.box import Box

class Rapid:
    pass

class ClassificationRapid(Rapid):
    """
    A classification rapid. Used as a multiple choice question for the labeler to answer.
    
    
    Args: 
        instruction (str): The instruction how to choose the options.
        answer_options (list[str]): The options that the labeler can choose from.
        truths (list[str]): The correct answers to the question.
        asset (MediaAsset | TextAsset): The asset that the labeler will be labeling.
        metadata (Sequence[Metadata]): The metadata that is attached to the rapid.
    """

    def __init__(self, instruction: str, answer_options: list[str], truths: list[str], asset: MediaAsset | TextAsset, metadata: Sequence[Metadata]):
        self.instruction = instruction
        self.answer_options = answer_options
        self.truths = truths
        self.asset = asset
        self.metadata = metadata

class CompareRapid(Rapid):
    """
    Used as a comparison of two assets for the labeler to compare.
    
    Args:
        instruction (str): The instruction that the labeler will be comparing the assets on.
        truth (str): The correct answer to the comparison. (has to be one of the assets)
        asset (MultiAsset): The assets that the labeler will be comparing.
        metadata (Sequence[Metadata]): The metadata that is attached to the rapid.
    """
    def __init__(self, instruction: str, truth: str, asset: MultiAsset, metadata: Sequence[Metadata]):
        self.instruction = instruction
        self.asset = asset
        self.truth = truth
        self.metadata = metadata

class SelectWordsRapid(Rapid):
    """
    Used to give the labeler a text and have them select words from it.
    
    Args:
        instruction (str): The instruction for the labeler.
        truths (list[int]): The indices of the words that are the correct answers.
        asset (MediaAsset): The asset that the labeler will be selecting words from.
        sentence (str): The sentence that the labeler will be selecting words from. (split up by spaces)
        strict_grading (bool): Whether the grading should be strict or not. 
            True means that all correct words and no wrong words have to be selected for the rapid to be marked as correct.
            False means that at least one correct word and no wrong words have to be selected for the rapid to be marked as correct.
    """
    def __init__(self, instruction: str, truths: list[int], asset: MediaAsset, sentence: str, strict_grading: bool):
        self.instruction = instruction
        self.truths = truths
        self.asset = asset
        self.sentence = sentence
        self.strict_grading = strict_grading

class LocateRapid(Rapid):
    """
    Used to have the labeler locate a specific object in an image.
    
    Args:
        instruction (str): The instructions on what the labeler should do.
        truths (list[Box]): The boxes that the object is located in.
        asset (MediaAsset): The image that the labeler is locating the object in.
        metadata (Sequence[Metadata]): The metadata that is attached to the rapid.    
    """
    def __init__(self, instruction: str, truths: list[Box], asset: MediaAsset, metadata: Sequence[Metadata]):
        self.instruction = instruction
        self.asset = asset
        self.truths = truths
        self.metadata = metadata

class DrawRapid(Rapid):
    """
    Used to have the labeler draw a specific object in an image.
    
    Args:
        instruction (str): The instructions on what the labeler should do.
        truths (list[Box]): The boxes that the object is located in.
        asset (MediaAsset): The image that the labeler is drawing the object in.
        metadata (Sequence[Metadata]): The metadata that is attached to the rapid.
    """
    def __init__(self, instruction: str, truths: list[Box], asset: MediaAsset, metadata: Sequence[Metadata]):
        self.instruction = instruction
        self.asset = asset
        self.truths = truths



from typing import Sequence
from urllib3._collections import HTTPHeaderDict

from rapidata.service.openapi_service import OpenAPIService
from rapidata.rapidata_client.assets.data_type_enum import RapidataDataTypes
from rapidata.rapidata_client.assets import MediaAsset, TextAsset, MultiAsset
from rapidata.rapidata_client.order.rapidata_order import RapidataOrder
from rapidata.rapidata_client.order._rapidata_order_builder import RapidataOrderBuilder
from rapidata.rapidata_client.metadata import PromptMetadata, SelectWordsMetadata
from rapidata.rapidata_client.referee._naive_referee import NaiveReferee
from rapidata.rapidata_client.referee._early_stopping_referee import EarlyStoppingReferee
from rapidata.rapidata_client.selection._base_selection import RapidataSelection
from rapidata.rapidata_client.selection.validation_selection import ValidationSelection
from rapidata.rapidata_client.selection.labeling_selection import LabelingSelection
from rapidata.rapidata_client.workflow import (
    Workflow,
    ClassifyWorkflow,
    CompareWorkflow,
    FreeTextWorkflow,
    SelectWordsWorkflow,
    LocateWorkflow,
    DrawWorkflow)
from rapidata.rapidata_client.selection.validation_selection import ValidationSelection
from rapidata.rapidata_client.selection.labeling_selection import LabelingSelection
from rapidata.rapidata_client.assets import MediaAsset, TextAsset, MultiAsset
from rapidata.rapidata_client.filter import RapidataFilter
from rapidata.rapidata_client.filter.rapidata_filters import RapidataFilters
from rapidata.rapidata_client.settings import RapidataSettings, RapidataSetting
from rapidata.rapidata_client.selection.rapidata_selections import RapidataSelections

from rapidata.api_client.exceptions import BadRequestException
from rapidata.api_client.models.query_model import QueryModel
from rapidata.api_client.models.page_info import PageInfo
from rapidata.api_client.models.root_filter import RootFilter
from rapidata.api_client.models.filter import Filter
from rapidata.api_client.models.sort_criterion import SortCriterion


class RapidataOrderManager:
    """
    Handels everything regarding the orders from creation to retrieval.
    
    Attributes:
        filters (RapidataFilters): The RapidataFilters instance.
        settings (RapidataSettings): The RapidataSettings instance.
        selections (RapidataSelections): The RapidataSelections instance."""

    def __init__(self, openapi_service: OpenAPIService):
        self._openapi_service = openapi_service
        self.filters = RapidataFilters
        self.settings = RapidataSettings
        self.selections = RapidataSelections
        self.__priority = 50

    def __get_selections(self, validation_set_id: str | None, labeling_amount=3) -> Sequence[RapidataSelection]:
        if validation_set_id:
            return [ValidationSelection(validation_set_id=validation_set_id), LabelingSelection(amount=labeling_amount-1)]
        return [LabelingSelection(amount=labeling_amount)]
    
    def __create_general_order(self,
            name: str,
            workflow: Workflow,
            assets: list[MediaAsset] | list[TextAsset] | list[MultiAsset],
            data_type: str = RapidataDataTypes.MEDIA,
            responses_per_datapoint: int = 10,
            contexts: list[str] | None = None,
            validation_set_id: str | None = None,
            confidence_threshold: float | None = None,
            filters: Sequence[RapidataFilter] = [],
            settings: Sequence[RapidataSetting] = [],
            sentences: list[str] | None = None,
            selections: Sequence[RapidataSelection] | None = None,
            default_labeling_amount: int = 3
        ) -> RapidataOrder:
        
        if contexts and len(contexts) != len(assets):
            raise ValueError("Number of contexts must match number of datapoints")
        
        if sentences and len(sentences) != len(assets):
            raise ValueError("Number of sentences must match number of datapoints")
        
        if sentences and contexts:
            raise ValueError("You can only use contexts or sentences, not both")
        
        if contexts and data_type == RapidataDataTypes.TEXT:
            print("Warning: Contexts are not supported for text data type. Ignoring contexts.")

        if not confidence_threshold:
            referee = NaiveReferee(responses=responses_per_datapoint)
        else:
            referee = EarlyStoppingReferee(
                threshold=confidence_threshold,
                max_vote_count=responses_per_datapoint,
            )

        order_builder = RapidataOrderBuilder(name=name, openapi_service=self._openapi_service)

        if selections and validation_set_id:
            print("Warning: You provided both selections and validation_set_id. Ignoring validation_set_id.")
        
        if selections is None:
            selections = self.__get_selections(validation_set_id, labeling_amount=default_labeling_amount)

        prompts_metadata = [PromptMetadata(prompt=prompt) for prompt in contexts] if contexts else None
        sentence_metadata = [SelectWordsMetadata(select_words=sentence) for sentence in sentences] if sentences else None

        metadata = prompts_metadata or sentence_metadata or None

        order = (order_builder
                 ._workflow(workflow)
                 ._media(
                     asset=assets,
                     metadata=metadata
                     )
                 ._referee(referee)
                 ._filters(filters)
                 ._selections(selections) 
                 ._settings(settings)
                 ._priority(self.__priority)
                 ._create()
                 )
        return order
    
    def _set_priority(self, priority: int):
        self.__priority = priority
        
    def create_classification_order(self,
            name: str,
            instruction: str,
            answer_options: list[str],
            datapoints: list[str],
            data_type: str = RapidataDataTypes.MEDIA,
            responses_per_datapoint: int = 10,
            contexts: list[str] | None = None,
            validation_set_id: str | None = None,
            confidence_threshold: float | None = None,
            filters: Sequence[RapidataFilter] = [],
            settings: Sequence[RapidataSetting] = [],
            selections: Sequence[RapidataSelection] | None = None,
        ) -> RapidataOrder:
        """Create a classification order.
        
        Args:
            name (str): The name of the order. (Will not be shown to the labeler)
            instruction (str): The instruction for how the data should be classified.
            answer_options (list[str]): The list of options for the classification.
            datapoints (list[str]): The list of datapoints for the classification - each datapoint will be labeled.
            data_type (str, optional): The data type of the datapoints. Defaults to RapidataDataTypes.MEDIA. \n
                Other option: RapidataDataTypes.TEXT ("text").
            responses_per_datapoint (int, optional): The number of responses that will be collected per datapoint. Defaults to 10.
            contexts (list[str], optional): The list of contexts for the classification. Defaults to None.\n
                If provided has to be the same length as datapoints and will be shown in addition to the instruction and options. (Therefore will be different for each datapoint)
                Will be match up with the datapoints using the list index.
            validation_set_id (str, optional): The ID of the validation set. Defaults to None.\n
                If provided, one validation task will be shown infront of the datapoints that will be labeled.
            confidence_threshold (float, optional): The probability threshold for the classification. Defaults to None.\n
                If provided, the classification datapoint will stop after the threshold is reached or at the number of responses, whatever happens first.
            filters (Sequence[RapidataFilter], optional): The list of filters for the classification. Defaults to []. Decides who the tasks should be shown to.
            settings (Sequence[RapidataSetting], optional): The list of settings for the classification. Defaults to []. Decides how the tasks should be shown.
            selections (Sequence[RapidataSelection], optional): The list of selections for the classification. Defaults to None. Decides in what order the tasks should be shown.
        """
        
        if data_type == RapidataDataTypes.MEDIA:
            assets = [MediaAsset(path=path) for path in datapoints]
        elif data_type == RapidataDataTypes.TEXT:
            assets = [TextAsset(text=text) for text in datapoints]
        else:
            raise ValueError(f"Unsupported data type: {data_type}, must be one of {RapidataDataTypes._possible_values()}")
        
        return self.__create_general_order(
            name=name,
            workflow=ClassifyWorkflow(
                instruction=instruction,
                answer_options=answer_options
            ),
            assets=assets,
            data_type=data_type,
            responses_per_datapoint=responses_per_datapoint,
            contexts=contexts,
            validation_set_id=validation_set_id,
            confidence_threshold=confidence_threshold,
            filters=filters,
            selections=selections,
            settings=settings
        )
    
    def create_compare_order(self,
            name: str,
            instruction: str,
            datapoints: list[list[str]],
            data_type: str = RapidataDataTypes.MEDIA,
            responses_per_datapoint: int = 10,
            contexts: list[str] | None = None,
            validation_set_id: str | None = None,
            confidence_threshold: float | None = None,
            filters: Sequence[RapidataFilter] = [],
            settings: Sequence[RapidataSetting] = [],
            selections: Sequence[RapidataSelection] | None = None,
        ) -> RapidataOrder:
        """Create a compare order.

        Args:
            name (str): The name of the order. (Will not be shown to the labeler)
            instruction (str): The instruction for the comparison. Will be shown along side each datapoint.
            datapoints (list[list[str]]): Outher list is the datapoints, inner list is the options for the comparison - each datapoint will be labeled.
            data_type (str, optional): The data type of the datapoints. Defaults to RapidataDataTypes.MEDIA. \n
                Other option: RapidataDataTypes.TEXT ("text").
            responses_per_datapoint (int, optional): The number of responses that will be collected per datapoint. Defaults to 10.
            contexts (list[str], optional): The list of contexts for the comparison. Defaults to None.\n
                If provided has to be the same length as datapoints and will be shown in addition to the instruction. (Therefore will be different for each datapoint)
                Will be match up with the datapoints using the list index.
            validation_set_id (str, optional): The ID of the validation set. Defaults to None.\n
                If provided, one validation task will be shown infront of the datapoints that will be labeled.
            confidence_threshold (float, optional): The probability threshold for the comparison. Defaults to None.\n
                If provided, the comparison datapoint will stop after the threshold is reached or at the number of responses, whatever happens first.
            filters (Sequence[RapidataFilter], optional): The list of filters for the comparison. Defaults to []. Decides who the tasks should be shown to.
            settings (Sequence[RapidataSetting], optional): The list of settings for the comparison. Defaults to []. Decides how the tasks should be shown.
            selections (Sequence[RapidataSelection], optional): The list of selections for the comparison. Defaults to None. Decides in what order the tasks should be shown.
        """

        if data_type == RapidataDataTypes.MEDIA:
            assets = [MultiAsset([MediaAsset(path=path) for path in datapoint]) for datapoint in datapoints]
        elif data_type == RapidataDataTypes.TEXT:
            assets = [MultiAsset([TextAsset(text=text) for text in datapoint]) for datapoint in datapoints]
        else:
            raise ValueError(f"Unsupported data type: {data_type}, must be one of {RapidataDataTypes._possible_values()}")
        
        return self.__create_general_order(
            name=name,
            workflow=CompareWorkflow(
                instruction=instruction
            ),
            assets=assets,
            data_type=data_type,
            responses_per_datapoint=responses_per_datapoint,
            contexts=contexts,
            validation_set_id=validation_set_id,
            confidence_threshold=confidence_threshold,
            filters=filters,
            selections=selections,
            settings=settings
        )
    
    def create_free_text_order(self,
            name: str,
            instruction: str,
            datapoints: list[str],
            data_type: str = RapidataDataTypes.MEDIA,
            responses_per_datapoint: int = 10,
            filters: Sequence[RapidataFilter] = [],
            settings: Sequence[RapidataSetting] = [],
            selections: Sequence[RapidataSelection] | None = None,
        ) -> RapidataOrder:
        """Create a free text order.

        Args:
            name (str): The name of the order.
            instruction (str): The instruction to answer with free text. Will be shown along side each datapoint.
            datapoints (list[str]): The list of datapoints for the free text - each datapoint will be labeled.
            data_type (str, optional): The data type of the datapoints. Defaults to RapidataDataTypes.MEDIA. \n
                Other option: RapidataDataTypes.TEXT ("text").
            responses_per_datapoint (int, optional): The number of responses that will be collected per datapoint. Defaults to 10.
            filters (Sequence[RapidataFilter], optional): The list of filters for the free text. Defaults to []. Decides who the tasks should be shown to.
            settings (Sequence[RapidataSetting], optional): The list of settings for the free text. Defaults to []. Decides how the tasks should be shown.
            selections (Sequence[RapidataSelection], optional): The list of selections for the free text. Defaults to None. Decides in what order the tasks should be shown.
        """

        if data_type == RapidataDataTypes.MEDIA:
            assets = [MediaAsset(path=path) for path in datapoints]
        elif data_type == RapidataDataTypes.TEXT:
            assets = [TextAsset(text=text) for text in datapoints]
        else:
            raise ValueError(f"Unsupported data type: {data_type}, must be one of {RapidataDataTypes._possible_values()}")

        return self.__create_general_order(
            name=name,
            workflow=FreeTextWorkflow(
                instruction=instruction
            ),
            assets=assets,
            data_type=data_type,
            responses_per_datapoint=responses_per_datapoint,
            filters=filters,
            selections=selections,
            settings=settings,
            default_labeling_amount=1
        )
    
    def create_select_words_order(self,
            name: str,
            instruction: str,
            datapoints: list[str],
            sentences: list[str],
            responses_per_datapoint: int = 10,
            validation_set_id: str | None = None,
            filters: Sequence[RapidataFilter] = [],
            settings: Sequence[RapidataSetting] = [],
            selections: Sequence[RapidataSelection] | None = None,
        ) -> RapidataOrder:
        """Create a select words order.

        Args:
            name (str): The name of the order.
            instruction (str): The instruction for how the words should be selected. Will be shown along side each datapoint.
            datapoints (list[str]): The list of datapoints for the select words - each datapoint will be labeled.
            sentences (list[str]): The list of sentences for the select words - Will be split up by spaces and shown along side each datapoint.\n
                Must be the same length as datapoints.
            responses_per_datapoint (int, optional): The number of responses that will be collected per datapoint. Defaults to 10.
            validation_set_id (str, optional): The ID of the validation set. Defaults to None.\n
                If provided, one validation task will be shown infront of the datapoints that will be labeled.
            filters (Sequence[RapidataFilter], optional): The list of filters for the select words. Defaults to []. Decides who the tasks should be shown to.
            settings (Sequence[RapidataSetting], optional): The list of settings for the select words. Defaults to []. Decides how the tasks should be shown.
            selections (Sequence[RapidataSelection], optional): The list of selections for the select words. Defaults to None. Decides in what order the tasks should be shown.
        """

        assets = [MediaAsset(path=path) for path in datapoints]
        
        return self.__create_general_order(
            name=name,
            workflow=SelectWordsWorkflow(
                instruction=instruction
            ),
            assets=assets,
            responses_per_datapoint=responses_per_datapoint,
            validation_set_id=validation_set_id,
            filters=filters,
            selections=selections,
            settings=settings,
            sentences=sentences,
            default_labeling_amount=2
        )
    
    def create_locate_order(self,
            name: str,
            target: str,
            datapoints: list[str],
            responses_per_datapoint: int = 10,
            contexts: list[str] | None = None,
            validation_set_id: str | None = None,
            filters: Sequence[RapidataFilter] = [],
            settings: Sequence[RapidataSetting] = [],
            selections: Sequence[RapidataSelection] | None = None,
        ) -> RapidataOrder:
        """Create a locate order.

        Args:
            name (str): The name of the order.
            target (str): The target what should be located. Will be shown along side each datapoint.
            datapoints (list[str]): The list of datapoints for the locate - each datapoint will be labeled.
            responses_per_datapoint (int, optional): The number of responses that will be collected per datapoint. Defaults to 10.
            contexts (list[str], optional): The list of contexts for the comparison. Defaults to None.\n
                If provided has to be the same length as datapoints and will be shown in addition to the instruction. (Therefore will be different for each datapoint)
                Will be match up with the datapoints using the list index.
            validation_set_id (str, optional): The ID of the validation set. Defaults to None.\n
                If provided, one validation task will be shown infront of the datapoints that will be labeled.
            filters (Sequence[RapidataFilter], optional): The list of filters for the locate. Defaults to []. Decides who the tasks should be shown to.
            settings (Sequence[RapidataSetting], optional): The list of settings for the locate. Defaults to []. Decides how the tasks should be shown.
            selections (Sequence[RapidataSelection], optional): The list of selections for the locate. Defaults to None. Decides in what order the tasks should be shown.
        """

        assets = [MediaAsset(path=path) for path in datapoints]

        return self.__create_general_order(
            name=name,
            workflow=LocateWorkflow(target=target),
            assets=assets,
            responses_per_datapoint=responses_per_datapoint,
            contexts=contexts,
            validation_set_id=validation_set_id,
            filters=filters,
            selections=selections,
            settings=settings
        )

    def create_draw_order(self,
            name: str,
            target: str,
            datapoints: list[str],
            responses_per_datapoint: int = 10,
            contexts: list[str] | None = None,
            validation_set_id: str | None = None,
            filters: Sequence[RapidataFilter] = [],
            settings: Sequence[RapidataSetting] = [],
            selections: Sequence[RapidataSelection] | None = None,
        ) -> RapidataOrder:
        """Create a draw order.

        Args:
            name (str): The name of the order.
            target (str): The target for how the lines should be drawn. Will be shown along side each datapoint.
            datapoints (list[str]): The list of datapoints for the draw lines - each datapoint will be labeled.
            responses_per_datapoint (int, optional): The number of responses that will be collected per datapoint. Defaults to 10.
            contexts (list[str], optional): The list of contexts for the comparison. Defaults to None.\n
                If provided has to be the same length as datapoints and will be shown in addition to the instruction. (Therefore will be different for each datapoint)
                Will be match up with the datapoints using the list index.
            validation_set_id (str, optional): The ID of the validation set. Defaults to None.\n
                If provided, one validation task will be shown infront of the datapoints that will be labeled.
            filters (Sequence[RapidataFilter], optional): The list of filters for the draw lines. Defaults to []. Decides who the tasks should be shown to.
            settings (Sequence[RapidataSetting], optional): The list of settings for the draw lines. Defaults to []. Decides how the tasks should be shown.
            selections (Sequence[RapidataSelection], optional): The list of selections for the draw lines. Defaults to None. Decides in what order the tasks should be shown.
        """

        assets = [MediaAsset(path=path) for path in datapoints]

        return self.__create_general_order(
            name=name,
            workflow=DrawWorkflow(target=target),
            assets=assets,
            responses_per_datapoint=responses_per_datapoint,
            contexts=contexts,
            validation_set_id=validation_set_id,
            filters=filters,
            selections=selections,
            settings=settings
        )

    def get_order_by_id(self, order_id: str) -> RapidataOrder:
        """Get an order by ID.

        Args:
            order_id (str): The ID of the order.

        Returns:
            RapidataOrder: The Order instance.
        """

        try:
            order = self._openapi_service.order_api.order_get_by_id_get(order_id)
        except Exception:
            raise ValueError(f"Order with ID {order_id} not found.")

        return RapidataOrder(
            order_id=order_id, 
            name=order.order_name,
            openapi_service=self._openapi_service)

    def find_orders(self, name: str = "", amount: int = 1) -> list[RapidataOrder]:
        """Find your recent orders given criteria. If nothing is provided, it will return the most recent order.

        Args:
            name (str, optional): The name of the order - matching order will contain the name. Defaults to "" for any order.
            amount (int, optional): The amount of orders to return. Defaults to 1.

        Returns:
            list[RapidataOrder]: A list of RapidataOrder instances.
        """
        try:
            order_page_result = self._openapi_service.order_api.order_query_get(QueryModel(
                page=PageInfo(index=1, size=amount),
                filter=RootFilter(filters=[Filter(field="OrderName", operator="Contains", value=name)]),
                sortCriteria=[SortCriterion(direction="Desc", propertyName="OrderDate")]
                ))

        except BadRequestException as e:
            raise ValueError(f"Error occured during request. \nError: {e.body} \nTraceid: {e.headers.get('X-Trace-Id') if isinstance(e.headers, HTTPHeaderDict) else 'Unknown'}")

        except Exception as e:
            raise ValueError(f"Unknown error occured: {e}")

        orders = [self.get_order_by_id(order.id) for order in order_page_result.items]
        return orders



from enum import Enum
from rapidata.api_client.models.gender import Gender as BackendGenderModel

class Gender(Enum):
    """Gender Enum
    
    Represents the gender of a user. Used to filter who to target based on genders.
    
    Attributes:
        MALE (Gender): Represents the Male gender.
        FEMALE (Gender): Represents the Female gender.
        OTHER (Gender): Represents any other gender.
    """
    MALE = BackendGenderModel.MALE
    FEMALE = BackendGenderModel.FEMALE
    OTHER = BackendGenderModel.OTHER

    def _to_backend_model(self) -> BackendGenderModel:
        return BackendGenderModel(self.value)



import os
from rapidata.api_client.models.attach_category_truth import AttachCategoryTruth
from rapidata.api_client.models.classify_payload import ClassifyPayload
from rapidata.api_client.models.compare_payload import ComparePayload
from rapidata.api_client.models.compare_truth import CompareTruth
from rapidata.api_client.models.transcription_payload import TranscriptionPayload
from rapidata.api_client.models.transcription_truth import TranscriptionTruth
from rapidata.api_client.models.transcription_word import TranscriptionWord
from rapidata.api_client.models.locate_payload import LocatePayload
from rapidata.api_client.models.locate_box_truth import LocateBoxTruth
from rapidata.api_client.models.line_payload import LinePayload
from rapidata.api_client.models.bounding_box_truth import BoundingBoxTruth
from rapidata.api_client.models.box_shape import BoxShape
from rapidata.rapidata_client.validation.rapidata_validation_set import (
    RapidataValidationSet,
)
from rapidata.rapidata_client.assets import MediaAsset, TextAsset, MultiAsset
from rapidata.rapidata_client.validation._validation_rapid_parts import ValidatioRapidParts
from rapidata.rapidata_client.metadata._base_metadata import Metadata
from rapidata.service.openapi_service import OpenAPIService
from rapidata.rapidata_client.validation.rapids.box import Box

from rapidata.rapidata_client.validation.rapids.rapids import (
    Rapid, 
    ClassificationRapid,
    CompareRapid,
    SelectWordsRapid,
    LocateRapid,
    DrawRapid
)
from typing import Sequence


class ValidationSetBuilder:
    """The ValidationSetBuilder is used to build a validation set.
    Give the validation set a name and then add classify, compare, or transcription rapid parts to it.
    Get a `ValidationSetBuilder` by calling [`rapi.new_validation_set()`](../rapidata_client.md/#rapidata.rapidata_client.rapidata_client.RapidataClient.new_validation_set).

    Args:
        name (str): The name of the validation set.
        openapi_service (OpenAPIService): An instance of OpenAPIService to interact with the API.
    """

    def __init__(self, name: str, openapi_service: OpenAPIService):
        self.name = name
        self.openapi_service = openapi_service
        self.validation_set_id: str | None = None
        self._rapid_parts: list[ValidatioRapidParts] = []

    def _submit(self, print_confirmation: bool = True) -> RapidataValidationSet:
        """Create the validation set by executing all HTTP requests. This should be the last method called on the builder.

        Returns:
            RapidataValidationSet: A RapidataValidationSet instance.

        Raises:
            ValueError: If the validation set creation fails.
        """
        result = (
            self.openapi_service.validation_api.validation_create_validation_set_post(
                name=self.name
            )
        )
        self.validation_set_id = result.validation_set_id

        if self.validation_set_id is None:
            raise ValueError("Failed to create validation set")

        validation_set = RapidataValidationSet(
            validation_set_id=self.validation_set_id,
            openapi_service=self.openapi_service,
            name=self.name,
        )

        for rapid_part in self._rapid_parts:
            validation_set._add_general_validation_rapid(
                payload=rapid_part.payload,
                truths=rapid_part.truths,
                metadata=rapid_part.metadata,
                asset=rapid_part.asset,
                randomCorrectProbability=rapid_part.randomCorrectProbability,
            )

        if print_confirmation:
            print(f"Validation set '{self.name}' created with ID {self.validation_set_id}")
            
        return validation_set
    
    def _add_rapid(self, rapid: Rapid):
        """Add a rapid to the validation set.
        To create the Rapid, use the RapidataClient.rapid_builder instance.

        Args:
            rapid (Rapid): The rapid to add to the validation set.
        """
        if not isinstance(rapid, Rapid):
            raise ValueError("This method only accepts Rapid instances")
        
        elif isinstance(rapid, ClassificationRapid):
            self.__add_classify_rapid(rapid.asset, rapid.instruction, rapid.answer_options, rapid.truths, rapid.metadata)

        elif isinstance(rapid, CompareRapid):
            self.__add_compare_rapid(rapid.asset, rapid.instruction, rapid.truth, rapid.metadata)

        elif isinstance(rapid, SelectWordsRapid):
            self.__add_select_words_rapid(rapid.asset, rapid.instruction, rapid.sentence, rapid.truths, rapid.strict_grading)
        
        elif isinstance(rapid, LocateRapid):
            self.__add_locate_rapid(rapid.asset, rapid.instruction, rapid.truths)

        elif isinstance(rapid, DrawRapid):
            self.__add_draw_rapid(rapid.asset, rapid.instruction, rapid.truths)

        else:
            raise ValueError("Unsupported rapid type")

        return self
    
    def __add_classify_rapid(
        self,
        asset: MediaAsset | TextAsset,
        instruction: str,
        answer_options: list[str],
        truths: list[str],
        metadata: Sequence[Metadata] = [],
    ):
        """Add a classify rapid to the validation set.

        Args:
            asset (MediaAsset | TextAsset): The asset for the rapid.
            instruction (str): The instruction for the rapid.
            answer_options (list[str]): The list of answer_options for the rapid.
            truths (list[str]): The list of truths for the rapid.
            metadata (Sequence[Metadata], optional): The metadata for the rapid. Defaults to an empty list.

        Returns:
            ValidationSetBuilder: The ValidationSetBuilder instance.

        Raises:
            ValueError: If the lengths of categories and truths are inconsistent.
        """
        if not all(truth in answer_options for truth in truths):
            raise ValueError("Truths must be part of the answer options")

        payload = ClassifyPayload(
            _t="ClassifyPayload", possibleCategories=answer_options, title=instruction
        )
        model_truth = AttachCategoryTruth(
            correctCategories=truths, _t="AttachCategoryTruth"
        )

        self._rapid_parts.append(
            ValidatioRapidParts(
                instruction=instruction,
                payload=payload,
                truths=model_truth,
                metadata=metadata,
                randomCorrectProbability=len(truths) / len(answer_options),
                asset=asset,
            )
        )
    
    def __add_compare_rapid(
        self,
        asset: MultiAsset,
        instruction: str,
        truth: str,
        metadata: Sequence[Metadata] = [],
    ):
        """Add a compare rapid to the validation set.

        Args:
            asset (MultiAsset): The assets for the rapid.
            instruction (str): The instruction for the comparison.
            truth (str): The truth identifier for the rapid.
            metadata (Sequence[Metadata], optional): The metadata for the rapid. Defaults to an empty list.

        Returns:
            ValidationSetBuilder: The ValidationSetBuilder instance.

        Raises:
            ValueError: If the number of assets is not exactly two.
        """
        payload = ComparePayload(_t="ComparePayload", criteria=instruction)
        # take only last part of truth path
        truth = os.path.basename(truth)
        model_truth = CompareTruth(_t="CompareTruth", winnerId=truth)

        if len(asset) != 2:
            raise ValueError("Compare rapid requires exactly two media paths")

        self._rapid_parts.append(
            ValidatioRapidParts(
                instruction=instruction,
                payload=payload,
                truths=model_truth,
                metadata=metadata,
                randomCorrectProbability=1 / len(asset),
                asset=asset,
            )
        )
    
    def __add_select_words_rapid(
        self,
        asset: MediaAsset | TextAsset,
        instruction: str,
        select_words: str,
        truths: list[int],
        strict_grading: bool | None = None,
        metadata: Sequence[Metadata] = [],
    ):
        """Add a select words rapid to the validation set.

        Args:
            asset (MediaAsset | TextAsset): The asset for the rapid.
            instruction (str): The instruction for the rapid.
            select words (list[str]): The select words for the rapid.
            truths (list[int]): The list of indices of the true word selections.
            strict_grading (bool | None, optional): The strict grading for the rapid. Defaults to None.
            metadata (Sequence[Metadata], optional): The metadata for the rapid.

        Returns:
            ValidationSetBuilder: The ValidationSetBuilder instance.

        Raises:
            ValueError: If a correct word is not found in the select words.
        """
        transcription_words = [
            TranscriptionWord(word=word, wordIndex=i)
            for i, word in enumerate(select_words.split())
        ]

        true_words = []
        for idx in truths:
            assert isinstance(idx, int), "truths must be a list of integers"
            if idx > len(transcription_words) - 1:
                raise ValueError(f"Index {idx} is out of bounds")
            true_words.append(transcription_words[idx])

        payload = TranscriptionPayload(
            _t="TranscriptionPayload", title=instruction, transcription=transcription_words
        )

        model_truth = TranscriptionTruth(
            _t="TranscriptionTruth",
            correctWords=true_words,
            strictGrading=strict_grading,
        )

        self._rapid_parts.append(
            ValidatioRapidParts(
                instruction=instruction,
                asset=asset,
                payload=payload,
                truths=model_truth,
                metadata=metadata,
                randomCorrectProbability = 1 / len(transcription_words),
            )
        )
    
    def __add_locate_rapid(
        self,
        asset: MediaAsset,
        instruction: str,
        truths: list[Box]
    ):
        """Add a locate rapid to the validation set.

        Args:
            instruction (str): The instruction for the locate rapid.
            asset (MediaAsset): The asset for the rapid.
            truths (list[Box]): The truths for the rapid.

        Returns:
            ValidationSetBuilder: The ValidationSetBuilder instance.
        """
        payload = LocatePayload(
            _t="LocatePayload", target=instruction
        )

        img_dimensions = asset.get_image_dimension()

        if not img_dimensions:
            raise ValueError("Failed to get image dimensions")

        model_truth = LocateBoxTruth(
            _t="LocateBoxTruth", 
            boundingBoxes=[BoxShape(
                _t="BoxShape",
                xMin=truth.x_min / img_dimensions[0] * 100,
                xMax=truth.x_max / img_dimensions[0] * 100,
                yMax=truth.y_max / img_dimensions[1] * 100,
                yMin=truth.y_min / img_dimensions[1] * 100,
            ) for truth in truths]
        )

        coverage = self._calculate_boxes_coverage(truths, img_dimensions[0], img_dimensions[1])

        self._rapid_parts.append(
            ValidatioRapidParts(
                instruction=instruction,
                payload=payload,
                truths=model_truth,
                metadata=[],
                randomCorrectProbability=coverage,
                asset=asset,
            )
        )

    def __add_draw_rapid(
        self,
        asset: MediaAsset,
        instruction: str,
        truths: list[Box]
    ):
        """Add a draw rapid to the validation set.

        Args:
            instruction (str): The instruction for the draw rapid.
            asset (MediaAsset): The asset for the rapid.
            truths (list[Box]): The truths for the rapid.

        Returns:
            ValidationSetBuilder: The ValidationSetBuilder instance.
        """

        payload = LinePayload(
            _t="LinePayload", target=instruction
        )

        img_dimensions = asset.get_image_dimension()

        if not img_dimensions:
            raise ValueError("Failed to get image dimensions")

        model_truth = BoundingBoxTruth(
            _t="BoundingBoxTruth", 
            xMax=truths[0].x_max / img_dimensions[0],
            xMin=truths[0].x_min / img_dimensions[0],
            yMax=truths[0].y_max / img_dimensions[1],
            yMin=truths[0].y_min / img_dimensions[1],
        ) # TO BE CHANGED BEFORE MERGING

        coverage = self._calculate_boxes_coverage(truths, img_dimensions[0], img_dimensions[1])

        self._rapid_parts.append(
            ValidatioRapidParts(
                instruction=instruction,
                payload=payload,
                truths=model_truth,
                metadata=[],
                randomCorrectProbability=coverage,
                asset=asset,
            )
        )


    def _calculate_boxes_coverage(self, boxes: list[Box], image_width: int, image_height: int) -> float:
        if not boxes:
            return 0.0
            
        # Convert all coordinates to integers for pixel-wise coverage
        pixels = set()
        for box in boxes:
            for x in range(int(box.x_min), int(box.x_max + 1)):
                for y in range(int(box.y_min), int(box.y_max + 1)):
                    if 0 <= x < image_width and 0 <= y < image_height:
                        pixels.add((x,y))
                        
        total_covered = len(pixels)
        return total_covered / (image_width * image_height)




from rapidata.api_client.models.validation_selection import ValidationSelection as ValidationSelectionModel
from rapidata.rapidata_client.selection._base_selection import RapidataSelection


class ValidationSelection(RapidataSelection):
    """Validation selection class.

    Decides how many validation rapids you want to show per session.
    
    Args:
        validation_set_id (str): The id of the validation set to be used.
        amount (int): The amount of validation rapids that will be shown per session of this validation set.
    """

    def __init__(self, validation_set_id: str, amount: int = 1):
        self.validation_set_id = validation_set_id
        self.amount = amount

    def _to_model(self):
        return ValidationSelectionModel(_t="ValidationSelection", validationSetId=self.validation_set_id, amount=self.amount)
    


from abc import ABC, abstractmethod
from typing import Any

from rapidata.api_client.models.simple_workflow_model import SimpleWorkflowModel
from rapidata.api_client.models.evaluation_workflow_model import EvaluationWorkflowModel
from rapidata.api_client.models.compare_workflow_model import CompareWorkflowModel
from rapidata.rapidata_client.referee._base_referee import Referee


class Workflow(ABC):

    def __init__(self, type: str):
        self._type = type
        self._target_country_codes: list[str] = []

    def _to_dict(self) -> dict[str, Any]:
        return {
            "_t": self._type,
        }

    @abstractmethod
    def _to_model(
        self,
    ) -> SimpleWorkflowModel | CompareWorkflowModel | EvaluationWorkflowModel:
        pass
